/*
 * tkb.c - Tinkerbell Kernel Module
 *
 * Copyright (C) 2013 Intel Corporation
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Author: Yoshi Wang <yoshix.wang@intel.com>
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/cpufreq.h>
#include <linux/pm_qos.h>
#include <linux/sched.h>
#include <linux/tkb.h>
#include <asm/uaccess.h>
#include <asm/intel_mid_pcihelpers.h>

enum {
	DEBUG_TKB = 1U << 0,
};
static int debug_mask;
module_param_named(debug_mask, debug_mask, int, S_IRUGO | S_IWUSR | S_IWGRP);

#define TKB_NAME "tkb"
#define TKB_IOCTL_MAGIC 'x'
#define TKB_IOCTL_CPUFREQ_SET		_IOW(TKB_IOCTL_MAGIC, 0x01, char *)
#define TKB_IOCTL_THERMAL_SET		_IOW(TKB_IOCTL_MAGIC, 0x02, char *)
#define TKB_IOCTL_STATUS_SHOW		_IOW(TKB_IOCTL_MAGIC, 0x03, char *)
#define TKB_IOCTL_MSGBUS_SET		_IOW(TKB_IOCTL_MAGIC, 0x04, char *)
#define TKB_IOCTL_MSGBUS_GET		_IOWR(TKB_IOCTL_MAGIC, 0x05, char *)
#define TKB_IOCTL_MSR_SET		_IOW(TKB_IOCTL_MAGIC, 0x06, char *)

#define MAX_THERMAL_LEVEL		10

#define TKB_THERMAL_STATE_INCREASE	1
#define TKB_THERMAL_STATE_DECREASE	2
#define TKB_THERMAL_FREQ_CHANGE		3

#define pr_tkb(args...) \
	do { \
		if (debug_mask & DEBUG_TKB) { \
			pr_info(TKB_NAME ": " args); \
		} \
	} while (0)

struct thermal_request {
	unsigned char levels;
	int temp[MAX_THERMAL_LEVEL];
	int min_freq[MAX_THERMAL_LEVEL];
	int max_freq[MAX_THERMAL_LEVEL];
};

struct tkb_info {
/*for cpufreq*/
	struct workqueue_struct *cpufreq_wq;
	struct work_struct cpufreq_work;
	struct pm_qos_request *cpufreq_req_h_freq;
	struct pm_qos_request *cpufreq_req_l_freq;
	int cpufreq_set_h_freq;
	int cpufreq_set_l_freq;
/*for thermal*/
	struct thermal_request t_req;
	int current_state;
	int current_temp;
	unsigned int default_h_freq;
	unsigned int default_l_freq;
	bool thermal_enable;
	struct pm_qos_request *thermal_req_l_freq;
	struct pm_qos_request *thermal_req_h_freq;
	int thermal_set_l_freq;
	int thermal_set_h_freq;
#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
	int current_high;
	int current_low;
	int thermal_service_low;
	int thermal_service_high;
	bool en_thermal_service_low;
	bool en_thermal_service_high;
	struct workqueue_struct *thermal_wq;
	struct work_struct set_threshold;
#endif
};

struct bit_request {
	unsigned int h;
	unsigned int l;
	unsigned int v;
};

#define IOSF_MBI_RD_MASK	0xFEFFFFFF
#define IOSF_MBI_WR_MASK	0X01000000

#define IOSF_MBI_MASK_HI	0xFFFFFF00
#define IOSF_MBI_MASK_LO	0x000000FF
#define IOSF_MBI_ENABLE		0xF0

#define IOSF_MBI_MAX_PORTS	0xFF

struct tkb_info *tkbi;

static int get_state_by_temp(int temp)
{
	int state = 0;

	for (state = 0; state < tkbi->t_req.levels; state++) {
		if (temp >= tkbi->t_req.temp[state])
			continue;
		else
			return state;
	}
	return state;
}

static void tkb_throttle_cpu_freq(int temp)
{
	int state;

	state = get_state_by_temp(temp);
	pr_tkb("state change from %d to %d\n", tkbi->current_state, state);
	tkbi->current_state = state;

	if (state == 0) {
		if (tkbi->thermal_set_h_freq != tkbi->default_h_freq ||
			tkbi->thermal_set_l_freq != tkbi->default_l_freq) {
			tkbi->thermal_set_h_freq = tkbi->default_h_freq;
			tkbi->thermal_set_l_freq = tkbi->default_l_freq;
			pm_qos_update_request(tkbi->thermal_req_h_freq,
					tkbi->thermal_set_h_freq);
			pm_qos_update_request(tkbi->thermal_req_l_freq,
					tkbi->thermal_set_l_freq);
			pr_tkb("throttle CPU freq (default), max_freq = %d, min_freq= %d\n",
					tkbi->thermal_set_h_freq, tkbi->thermal_set_l_freq);
		}
	} else if (state > 0) {
		if (tkbi->thermal_set_h_freq != tkbi->t_req.max_freq[state - 1] ||
				tkbi->thermal_set_l_freq != tkbi->t_req.min_freq[state - 1]) {
			tkbi->thermal_set_h_freq = tkbi->t_req.max_freq[state - 1];
			tkbi->thermal_set_l_freq = tkbi->t_req.min_freq[state - 1];

			if (tkbi->thermal_set_h_freq < 0)
				pm_qos_update_request(tkbi->thermal_req_h_freq,
					tkbi->default_h_freq);
			else
				pm_qos_update_request(tkbi->thermal_req_h_freq,
					tkbi->thermal_set_h_freq);

			if (tkbi->thermal_set_l_freq < 0)
				pm_qos_update_request(tkbi->thermal_req_l_freq,
					tkbi->default_l_freq);
			else
				pm_qos_update_request(tkbi->thermal_req_l_freq,
					tkbi->thermal_set_l_freq);

			pr_tkb("throttle CPU freq (%d), max_freq = %d, min_freq= %d\n",
					(state-1) , tkbi->thermal_set_h_freq,
					tkbi->thermal_set_l_freq);
		}
	}
}

#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
static int find_next_low_threshold(void)
{
	if (!tkbi->thermal_enable || tkbi->current_state == 0 ||
		(tkbi->thermal_service_low >= tkbi->t_req.temp[tkbi->current_state - 1]
		&& tkbi->thermal_service_low < tkbi->current_temp)) {
		tkbi->en_thermal_service_low = true;
		tkbi->current_low = tkbi->thermal_service_low;
		pr_tkb("Next low threshold is set by thermal service, threshold = %d\n",
			tkbi->thermal_service_low);
		return tkbi->thermal_service_low;
	} else {
		tkbi->en_thermal_service_low = false;
		tkbi->current_low = tkbi->t_req.temp[tkbi->current_state - 1];
		pr_tkb("Next low threshold is set by tkb, threshold = %d\n",
			tkbi->t_req.temp[tkbi->current_state - 1]);
		return tkbi->t_req.temp[tkbi->current_state - 1];
	}
}

static int find_next_high_threshold(void)
{
	if (!tkbi->thermal_enable || tkbi->current_state == tkbi->t_req.levels ||
		(tkbi->thermal_service_high <= tkbi->t_req.temp[tkbi->current_state]
		&& tkbi->thermal_service_high > tkbi->current_temp)) {
		tkbi->en_thermal_service_high = true;
		tkbi->current_high = tkbi->thermal_service_high;
		pr_tkb("Next high threshold is set by thermal service, threshold = %d\n",
			tkbi->thermal_service_high);
		return tkbi->thermal_service_high;
	} else {
		tkbi->en_thermal_service_high = false;
		tkbi->current_high = tkbi->t_req.temp[tkbi->current_state];
		pr_tkb("Next high threshold is set by tkb, threshold = %d\n",
				tkbi->t_req.temp[tkbi->current_state]);
		return tkbi->t_req.temp[tkbi->current_state];
	}
}

bool tkb_coretemp_trigger(int index, int temp, int event, int *l_thr,
	int *h_thr, int *is_enabled)
{
	int ret = true;

	pr_tkb("%s: index = %d, temp = %d, event = %d\n",
		__func__, index, temp, event);

	tkbi->current_temp = temp;

	if (!tkbi->thermal_enable) {
		*is_enabled = 0;
		return ret;
	} else {
		*is_enabled = 1;
	}

	/*temp2_input only*/
	if (index == 2) {
		tkb_throttle_cpu_freq(temp);

		if (event == 0 && tkbi->en_thermal_service_low) {
			/* trigger famework low thr */
			ret = true;
		} else if (event == 1 && tkbi->en_thermal_service_high) {
			/* trigger framework high thr. */
			ret = true;
		} else {
			/* triggered by tkb */
			ret = false;
		}

		*l_thr = find_next_low_threshold();
		*h_thr = find_next_high_threshold();
	} else {
		pr_tkb("%s: undemand index\n", __func__);
	}

	pr_tkb("%s: updated state = %d, low_thr = %d, high_thr = %d\n",
			 __func__, tkbi->current_state, *l_thr, *h_thr);

	return ret;
}

int tkb_get_coretemp_thr(int index, int val, int mask)
{
	pr_tkb("%s: index = %d, val = %d, mask = %d\n",
		__func__, index, val, mask);

	/*temp2_input only*/
	if (index == 2) {
		if (mask == THERM_MASK_THRESHOLD0) {
			tkbi->thermal_service_low = val;

			if (tkbi->thermal_enable)
				return find_next_low_threshold();

		} else if (mask == THERM_MASK_THRESHOLD1) {
			tkbi->thermal_service_high = val;

			if (tkbi->thermal_enable)
				return find_next_high_threshold();

		} else {
			pr_tkb("%s: unknown mask\n", __func__);
			return val;
		}
	} else {
		pr_tkb("%s: undemand index\n", __func__);
		return val;
	}

	return val;
}
#endif

void tkb_update_coretemp(int index, int temp)
{
	pr_tkb("%s: index = %d, temp = %d\n", __func__, index, temp);

	tkbi->current_temp = temp;

	if (!tkbi->thermal_enable)
		return;

	/*temp2_input only*/
	if (index == 2)
		tkb_throttle_cpu_freq(temp);
	else
		pr_tkb("%s: undemand index\n", __func__);
}

void tkb_print_thermal_thr(void)
{
	int i = 0;
	int levels = tkbi->t_req.levels;

	pr_tkb("levels = %d\n", levels);
	while (i < levels) {
		pr_tkb("Attr[%d]: threshold = %d, max_freq = %d, min_freq = %d\n", i,
			tkbi->t_req.temp[i], tkbi->t_req.max_freq[i],
			tkbi->t_req.min_freq[i]);
		i++;
	}
}

static void tkb_print_state(void)
{
	pr_tkb("--------tkb cpufreq status--------\n");
	pr_tkb("cpufreq_set_l_freq: %d\n", tkbi->cpufreq_set_l_freq);
	pr_tkb("cpufreq_set_h_freq: %d\n", tkbi->cpufreq_set_h_freq);
	pr_tkb("--------tkb thermal status--------\n");
	pr_tkb("enable: %d\n", tkbi->thermal_enable);
	if (tkbi->thermal_enable)
		tkb_print_thermal_thr();
	pr_tkb("current_state: %d\n", tkbi->current_state);
	pr_tkb("current_temp: %d\n", tkbi->current_temp);
	pr_tkb("thermal_set_l_freq: %d\n", tkbi->thermal_set_l_freq);
	pr_tkb("thermal_set_h_freq: %d\n", tkbi->thermal_set_h_freq);
	pr_tkb("default_l_freq: %d\n", tkbi->default_l_freq);
	pr_tkb("default_h_freq: %d\n", tkbi->default_h_freq);
#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
	pr_tkb("current_high: %d\n", tkbi->current_high);
	pr_tkb("current_low: %d\n", tkbi->current_low);
	pr_tkb("thermal_service_low: %d\n", tkbi->thermal_service_low);
	pr_tkb("thermal_service_high: %d\n", tkbi->thermal_service_high);
	pr_tkb("en_thermal_service_low: %s\n",
		tkbi->en_thermal_service_low ? "ture" : "false");
	pr_tkb("en_thermal_service_high: %s\n",
		tkbi->en_thermal_service_high ? "ture" : "false");
#endif
	pr_tkb("---------------------------------\n");
}

static void tkb_set_cpu_freq(struct work_struct *work)
{
	long old_prio;

	old_prio = task_nice(current);
	set_user_nice(current, -20);

	pr_tkb("%s\n", __func__);

	if (tkbi->cpufreq_set_h_freq >= 0) {
		pm_qos_update_request(tkbi->cpufreq_req_h_freq,
			tkbi->cpufreq_set_h_freq);
	} else {
			pm_qos_update_request(tkbi->cpufreq_req_h_freq,
				tkbi->default_h_freq);
	}

	if (tkbi->cpufreq_set_l_freq >= 0) {
		pm_qos_update_request(tkbi->cpufreq_req_l_freq,
			tkbi->cpufreq_set_l_freq);
	} else {
			pm_qos_update_request(tkbi->cpufreq_req_l_freq,
				tkbi->default_l_freq);
	}

	pr_tkb("%s end\n", __func__);
	set_user_nice(current, old_prio);

}

#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
static void tkb_set_thermal_threshold(struct work_struct *work)
{
	core_threshold_set(find_next_low_threshold(), find_next_high_threshold());
}
#endif

static int tkb_open(struct inode *inode, struct file *file)
{
	pr_debug("%s\n", __func__);

	return 0;
}

static int tkb_release(struct inode *inode, struct file *file)
{
	pr_debug("%s\n", __func__);

	return 0;
}

static int parse_cpufreq_request(char *cpufreq_request)
{
	char *pch = cpufreq_request;
	char *result;

	result = strsep(&pch, ",");
	if (result != NULL)
		tkbi->cpufreq_set_l_freq = simple_strtol(result, NULL, 10);
	else
		return -1;

	result = strsep(&pch, ",");
	if (result != NULL)
		tkbi->cpufreq_set_h_freq = simple_strtol(result, NULL, 10);
	else
		return -1;

	return 0;
}

static int parse_thermal_request(char *thermal_request)
{
	char *pch = thermal_request;
	int level_num = 0;
	char *result;

	result = strsep(&pch, ",");
	if (result != NULL)
		tkbi->t_req.levels = simple_strtol(result, NULL, 10);

	if (tkbi->t_req.levels == 0) {
		pr_tkb("%s: user disable thermal request", __func__);
		return -1;
	}

	if (tkbi->t_req.levels > MAX_THERMAL_LEVEL) {
		pr_tkb("%s: over level %d(%d)", __func__, tkbi->t_req.levels,
			MAX_THERMAL_LEVEL);
		return -1;
	}

	for (level_num; level_num < tkbi->t_req.levels; level_num++) {

		result = strsep(&pch, ",");
		if (result != NULL)
			tkbi->t_req.temp[level_num] = simple_strtol(result, NULL, 10);
		else
			break;

		result = strsep(&pch, ",");
		if (result != NULL)
			tkbi->t_req.min_freq[level_num] = simple_strtol(result, NULL, 10);
		else
			break;

		result = strsep(&pch, ",");
		if (result != NULL)
			tkbi->t_req.max_freq[level_num] = simple_strtol(result, NULL, 10);
		else
			break;
	}

	if (level_num != tkbi->t_req.levels || tkbi->t_req.levels < 0) {
		pr_tkb("%s: incorrect user configuration data format", __func__);
		return -1;
	}

	return 0;
}

static inline u32 iosf_mbi_form_mcr(u8 op, u8 port, u8 offset)
{
	return (op << 24) | (port << 16) | (offset << 8) | IOSF_MBI_ENABLE;
}

static u32 iosf_mbi_read(u8 port, u8 opcode, u32 offset)
{
	u32 mcr, mcrx;

	mcr = iosf_mbi_form_mcr(opcode, port, offset & IOSF_MBI_MASK_LO);
	mcrx = offset & IOSF_MBI_MASK_HI;

	return intel_mid_msgbus_read32_raw_ext(mcr & IOSF_MBI_RD_MASK, mcrx);
}

static void iosf_mbi_modify(u8 port, u8 opcode, u32 offset, u32 mdr, u32 mask)
{
	u32 mcr, mcrx, value;

	mcr = iosf_mbi_form_mcr(opcode, port, offset & IOSF_MBI_MASK_LO);
	mcrx = offset & IOSF_MBI_MASK_HI;

	/* Read current value */
	value = intel_mid_msgbus_read32_raw_ext(mcr & IOSF_MBI_RD_MASK, mcrx);

	/* Apply mask */
	value &= ~mask;
	mdr &= mask;
	value |= mdr;

	/* Write back */
	intel_mid_msgbus_write32_raw_ext(mcr | IOSF_MBI_WR_MASK, mcrx, value);
}

static void apply_msgbus_request(u8 port, u8 opcode, u32 offset,
		struct bit_request *bitset, int count)
{
	u32 smask = 0xffffffff, value = 0, mask = 0;
	unsigned int vh, vl;

	for (; count > 0 && bitset != NULL; count--) {
		vh = bitset->h;
		vl = bitset->l;

		if (vh < vl) {
			bitset->h = vl;
			bitset->l = vh;
		}

		if (bitset->h > 31 || bitset->l < 0) {
			pr_tkb("%s: invalid range %d:%d\n", __func__, bitset->h, bitset->l);
			bitset++;
			continue;
		}

		mask = (smask >> (31 - (bitset->h - bitset->l))) <<  bitset->l;
		value = (bitset->v << bitset->l) & mask;

		iosf_mbi_modify(port, opcode, offset, value, mask);
		bitset++;
	}
}

static void dump_bitset_request(u8 port, u8 opcode, u32 offset,
		struct bit_request *bitset, int count)
{
	pr_tkb("%s: port: 0x%02x, opcode: 0x%02x, offset: 0x%08x\n",
			__func__, port, opcode, offset);

	for (; count > 0 && bitset != NULL; count--) {
		pr_tkb("%s: %d:%d, 0x%08x\n", __func__,
				bitset->h, bitset->l, bitset->v);
		bitset++;
	}
}

/* expect input:
 * 0xPORT,0xOPCODE,0xOFFSET,h1:l1,0xVALUE,h2:l2,0xVALUE,...
 *
 *   unit	port	ReadOp	WriteOp
 *   ---------------------------------------------------------
 * - AUNIT	0x00	0x10	0x11	Private Reg access
 * - DUNIT	0x01	0x10	0x11	Prviate Reg access
 * - DUNIT1	0x07	0x10	0x11	Private Reg access
 * - GUNIT	0x06	0x00	0x01	Standard MMIO access
 * - PUNIT	0x04	0x06	0x07	Standard Private Reg access
 * - USB	0x43	0x06	0x07	Standard Private Reg access
 */
static int parse_msgbus_request(char *buff, unsigned readonly)
{
	char *pch = buff;
	char *result, *range, *v, *h, *l;
	struct bit_request bitset[32];
	int count = 0;
	u8 port = 0, opcode = 0;
	u32 offset = 0, value = 0;

	pr_tkb("%s: %s (%s)\n", __func__, pch,
			readonly ? "ReadOp" : "WriteOp");

	result = strsep(&pch, ",");
	if (result != NULL) {
		port = simple_strtol(result, NULL, 16) & IOSF_MBI_MAX_PORTS;
	} else {
		pr_tkb("%s: cannot extract port\n", __func__);
		return -EINVAL;
	}

	result = strsep(&pch, ",");
	if (result != NULL) {
		opcode = simple_strtol(result, NULL, 16);
	} else {
		pr_tkb("%s: cannot extract opcode\n", __func__);
		return -EINVAL;
	}

	result = strsep(&pch, ",");
	if (result != NULL) {
		offset = simple_strtol(result, NULL, 16);
	} else {
		pr_tkb("%s: cannot extract addr\n", __func__);
		return -EINVAL;
	}

	if (readonly) {
		value = iosf_mbi_read(port, opcode, offset);

		memset(buff, 0x00, sizeof(buff));

		sprintf(buff,
			"port: 0x%02x, opcode: 0x%02x, "
			"offset: 0x%08x value: 0x%08x",
			port, opcode, offset, value);

		pr_tkb("%s: %s\n", __func__, buff);
		return 0;
	}

	while (pch) {
		range = strsep(&pch, ",");

		if (range == NULL)
			goto invalid_range;

		h = strsep(&range, ":");
		l = range;

		if (h == NULL || l == NULL)
			goto invalid_range;

		bitset[count].h = simple_strtol(h, NULL, 10);
		bitset[count].l = simple_strtol(l, NULL, 10);

		v = strsep(&pch, ",");

		if (v == NULL)
			goto invalid_value;

		bitset[count].v = simple_strtol(v, NULL, 16);

		count++;
	}

	if (debug_mask & DEBUG_TKB)
		dump_bitset_request(port, opcode, offset, &bitset, count);

	apply_msgbus_request(port, opcode, offset, &bitset, count);

	return 0;

invalid_range:
	pr_tkb("%s: cannot extract range\n", __func__);
	return -EINVAL;
invalid_value:
	pr_tkb("%s: cannot extract value\n", __func__);
	return -EINVAL;
}

static void apply_msr_request(int cpu, u32 reg,
	struct bit_request *bitset, int count)
{
	u32 eax_temp_mask = 0, edx_temp_mask = 0, eax_temp = 0, edx_temp = 0;
	u32 eax_mask = 0, edx_mask = 0, eax_mod = 0, edx_mod = 0, eax, edx;
	unsigned int vh, vl;

	for (; count > 0 && bitset != NULL; count--) {
		eax_temp = edx_temp = eax_temp_mask = edx_temp_mask = 0;
		vh = bitset->h;
		vl = bitset->l;

		if (vh < vl) {
			bitset->h = vl;
			bitset->l = vh;
		}

		pr_tkb("%s: l=%d, h=%d\n", __func__, bitset->l, bitset->h);

		if (bitset->h > 63 || bitset->l < 0) {
			pr_tkb("%s: invalid range %d:%d\n", __func__, bitset->h, bitset->l);
			bitset++;
			continue;
		}

		if (bitset->h <= 31) {
			eax_temp_mask = (0xffffffff >> (31 - (bitset->h - bitset->l))) <<  bitset->l;
			pr_tkb("%s: mask=0x%x\n", __func__, eax_temp_mask);
			eax_temp = (bitset->v << bitset->l) & eax_temp_mask;
		} else if (bitset->l > 31) {
			bitset->l -= 32;
			bitset->h -= 32;
			edx_temp_mask = (0xffffffff >> (31 - (bitset->h - bitset->l))) <<  bitset->l;
			pr_tkb("%s: mask=0x%x\n", __func__, edx_temp_mask);
			edx_temp = (bitset->v << bitset->l) & edx_temp_mask;
		} else {
			eax_temp_mask = (0xffffffff >> (31 - (31 - bitset->l))) <<  bitset->l;
			pr_tkb("%s: maska=0x%x\n", __func__, eax_temp_mask);
			eax_temp = (bitset->v << bitset->l) & eax_temp_mask;
			bitset->h -= 32;
			edx_temp_mask = (0xffffffff >> (31 - (bitset->h - 0))) << 0;
			pr_tkb("%s: maskd=0x%x\n", __func__, edx_temp_mask);
			edx_temp = (bitset->v << 0) & edx_temp_mask;
		}

		pr_tkb("%s: eax_temp=0x%x, edx_temp=0x%x, "
			"eax_temp_mask=0x%x, edx_temp_mask=0x%x,\n",
			__func__, eax_temp, edx_temp, eax_temp_mask, edx_temp_mask);

		eax_mask |= eax_temp_mask;
		eax_mod |= eax_temp;
		edx_mask |= edx_temp_mask;
		edx_mod |= edx_temp;
		pr_tkb("%s: eax=0x%x, edx=0x%x, eax_mask=0x%x, edx_mask=0x%x\n",
			__func__, eax_mod, edx_mod, eax_mask, edx_mask);

		bitset++;
	}

	if (cpu != -1) {
		rdmsr_safe_on_cpu(cpu, reg, &eax, &edx);
		pr_tkb("%s: rdmsr cpu=%d, eax=0x%x, edx=0x%x\n", __func__, cpu, eax, edx);
		eax = (eax & ~eax_mask) | eax_mod;
		edx = (edx & ~edx_mask) | edx_mod;
		wrmsr_safe_on_cpu(cpu, reg, eax, edx);
		pr_tkb("%s: wrmsr cpu=%d, eax=0x%x, edx=0x%x\n", __func__, cpu, eax, edx);
	} else {
		for_each_online_cpu(cpu) {
			rdmsr_safe_on_cpu(cpu, reg, &eax, &edx);
			pr_tkb("%s: rdmsr cpu=%d, eax=0x%x, edx=0x%x\n", __func__, cpu, eax, edx);
			eax = (eax & ~eax_mask) | eax_mod;
			edx = (edx & ~edx_mask) | edx_mod;
			wrmsr_safe_on_cpu(cpu, reg, eax, edx);
			pr_tkb("%s: wrmsr cpu=%d, eax=0x%x, edx=0x%x\n", __func__, cpu, eax, edx);
		}
	}
}

static int parse_msr_request(char *buff)
{
	char *pch = buff;
	char *result, *range, *v, *h, *l;
	struct bit_request bitset[32];
	int count = 0;
	int cpu = 0;
	u32 reg;

	pr_tkb("%s: %s\n", __func__, pch);

	result = strsep(&pch, ",");
	if (result != NULL) {
		if (strncmp(result, "all", 3) == 0) {
			cpu = -1;
		} else {
			cpu = simple_strtol(result, NULL, 10);
		}
	} else {
		pr_tkb("%s: cannot extract cpu\n", __func__);
		return -EINVAL;
	}

	result = strsep(&pch, ",");
	if (result != NULL) {
		reg = simple_strtol(result, NULL, 16);
	} else {
		pr_tkb("%s: cannot extract reg\n", __func__);
		return -EINVAL;
	}

	while (pch) {
		range = strsep(&pch, ",");

		if (range == NULL)
			goto invalid_range;

		h = strsep(&range, ":");
		l = range;

		if (h == NULL || l == NULL)
			goto invalid_range;

		bitset[count].h = simple_strtol(h, NULL, 10);
		bitset[count].l = simple_strtol(l, NULL, 10);

		v = strsep(&pch, ",");

		if (v == NULL)
			goto invalid_value;

		bitset[count].v = simple_strtol(v, NULL, 16);

		count++;
	}

	if (debug_mask & DEBUG_TKB) {
		pr_info("%s: cpu=%d, reg=0x%x", __func__, cpu, reg);
		dump_bitset_request(0, 0, 0, &bitset, count);
	}

	apply_msr_request(cpu, reg, &bitset, count);

	return 0;

invalid_range:
	pr_tkb("%s: cannot extract range\n", __func__);
	return -EINVAL;
invalid_value:
	pr_tkb("%s: cannot extract value\n", __func__);
	return -EINVAL;
}
static long tkb_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	char buff[256];
	void __user *argp = (void __user *)arg;

	pr_tkb("%s : cmd:%d\n", __func__, cmd);

	switch (cmd) {
	case TKB_IOCTL_CPUFREQ_SET:
	{
		pr_tkb("receive TKB_IOCTL_CPUFREQ_SET cmd\n");
		if (copy_from_user(buff, argp, sizeof(buff)/sizeof(buff[0])))
			return -EFAULT;

		if (parse_cpufreq_request(buff))
			return -EFAULT;
		else
			queue_work(tkbi->cpufreq_wq, &tkbi->cpufreq_work);

		if (debug_mask & DEBUG_TKB)
			tkb_print_state();
		return 0;

	}
	case TKB_IOCTL_THERMAL_SET:
	{
		pr_tkb("receive TKB_IOCTL_THERMAL_SET cmd\n");

		if (copy_from_user(buff, argp, sizeof(buff)/sizeof(buff[0])))
			return -EFAULT;

		if (parse_thermal_request(buff) == 0) {
			tkbi->thermal_enable = true;
			tkbi->current_state = 0;
#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
			schedule_work_on(0, &tkbi->set_threshold);
#endif
		} else {
			tkbi->thermal_enable = false;
			tkbi->current_state = -1;
			tkbi->thermal_set_l_freq = tkbi->default_l_freq;
			tkbi->thermal_set_h_freq = tkbi->default_h_freq;
			pm_qos_update_request(tkbi->thermal_req_l_freq,
				tkbi->thermal_set_l_freq);
			pm_qos_update_request(tkbi->thermal_req_h_freq,
				tkbi->thermal_set_h_freq);
#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
			schedule_work_on(0, &tkbi->set_threshold);
#endif
		}

		if (debug_mask & DEBUG_TKB)
			tkb_print_state();
		return 0;
	}
	case TKB_IOCTL_STATUS_SHOW:
	{
		if (debug_mask & DEBUG_TKB)
			tkb_print_state();
		return 0;
	}
	case TKB_IOCTL_MSGBUS_SET:
	{
		if (copy_from_user(buff, argp, sizeof(buff)/sizeof(buff[0])))
			return -EFAULT;

		if (parse_msgbus_request(buff, 0))
			return -EINVAL;

		return 0;
	}
	case TKB_IOCTL_MSGBUS_GET:
	{
		if (copy_from_user(buff, argp, sizeof(buff)/sizeof(buff[0])))
			return -EFAULT;

		if (parse_msgbus_request(buff, 1))
			return -EINVAL;

		if (copy_to_user(argp, buff, sizeof(buff)/sizeof(buff[0])))
			return -EFAULT;

		return 0;
	}
	case TKB_IOCTL_MSR_SET:
	{
		if (copy_from_user(buff, argp, sizeof(buff)/sizeof(buff[0])))
			return -EFAULT;

		if (parse_msr_request(buff))
			return -EINVAL;

		return 0;
	}
	default:
		pr_err("%s: Wrong command: %d\n", __func__, cmd);
		return 0;
	}
}

static const struct file_operations tkb_fops = {
	.owner		= THIS_MODULE,
	.open		= tkb_open,
	.release	= tkb_release,
	.unlocked_ioctl	= tkb_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = tkb_ioctl,
#endif
};

static struct miscdevice tkb_misc = {
	.fops		= &tkb_fops,
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= TKB_NAME,
};

static int __init tkb_init(void)
{
	int rc = 0;

	pr_info("%s\n", __func__);

	tkbi = kzalloc(sizeof(struct tkb_info), GFP_KERNEL);
	tkbi->cpufreq_req_h_freq =
		kzalloc(sizeof(struct pm_qos_request), GFP_KERNEL);
	tkbi->cpufreq_req_l_freq =
		kzalloc(sizeof(struct pm_qos_request), GFP_KERNEL);
	tkbi->cpufreq_set_h_freq = PM_QOS_CPU_FREQ_MAX_DEFAULT_VALUE;
	tkbi->cpufreq_set_l_freq = PM_QOS_CPU_FREQ_MIN_DEFAULT_VALUE;

	pm_qos_add_request(tkbi->cpufreq_req_h_freq,
		PM_QOS_CPU_FREQ_MAX, PM_QOS_CPU_FREQ_MAX_DEFAULT_VALUE);
	pm_qos_add_request(tkbi->cpufreq_req_l_freq,
		PM_QOS_CPU_FREQ_MIN, PM_QOS_CPU_FREQ_MIN_DEFAULT_VALUE);

	if (!tkbi->cpufreq_wq) {
		tkbi->cpufreq_wq = create_freezable_workqueue(TKB_NAME);
		if (!tkbi->cpufreq_wq)
			goto error_wq;
	}

	INIT_WORK(&tkbi->cpufreq_work, tkb_set_cpu_freq);

#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
	INIT_WORK(&tkbi->set_threshold,
					tkb_set_thermal_threshold);
#endif

	tkbi->thermal_enable = false;
	tkbi->current_state = -1;
	tkbi->current_temp = -1;

#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
	tkbi->en_thermal_service_low = true;
	tkbi->en_thermal_service_high = true;
	tkbi->thermal_service_high = 90000;
	tkbi->thermal_service_low = 0;
#endif

	tkbi->default_h_freq = PM_QOS_CPU_FREQ_MAX_DEFAULT_VALUE;
	tkbi->default_l_freq = PM_QOS_CPU_FREQ_MIN_DEFAULT_VALUE;
	tkbi->thermal_req_h_freq =
		kzalloc(sizeof(struct pm_qos_request), GFP_KERNEL);
	tkbi->thermal_req_l_freq =
		kzalloc(sizeof(struct pm_qos_request), GFP_KERNEL);
	tkbi->thermal_set_h_freq = PM_QOS_CPU_FREQ_MAX_DEFAULT_VALUE;
	tkbi->thermal_set_l_freq = PM_QOS_CPU_FREQ_MIN_DEFAULT_VALUE;

	pm_qos_add_request(tkbi->thermal_req_h_freq,
		PM_QOS_CPU_FREQ_MAX, PM_QOS_CPU_FREQ_MAX_DEFAULT_VALUE);
	pm_qos_add_request(tkbi->thermal_req_l_freq,
		PM_QOS_CPU_FREQ_MIN, PM_QOS_CPU_FREQ_MIN_DEFAULT_VALUE);

	rc = misc_register(&tkb_misc);
	if (rc < 0)
		goto error_misc;

	return rc;

error_wq:
error_misc:
	pm_qos_remove_request(tkbi->cpufreq_req_l_freq);
	pm_qos_remove_request(tkbi->cpufreq_req_h_freq);
	pm_qos_remove_request(tkbi->thermal_req_l_freq);
	pm_qos_remove_request(tkbi->thermal_req_h_freq);
	kfree(tkbi->cpufreq_req_l_freq);
	kfree(tkbi->cpufreq_req_h_freq);
	kfree(tkbi->thermal_req_l_freq);
	kfree(tkbi->thermal_req_h_freq);
	cancel_work_sync(&tkbi->cpufreq_work);
#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
	cancel_work_sync(&tkbi->set_threshold);
#endif
	kfree(tkbi);

	return -1;
}
module_init(tkb_init);

static void __exit tkb_exit(void)
{
	pr_info("%s\n", __func__);

	if (tkbi->cpufreq_wq)
		destroy_workqueue(tkbi->cpufreq_wq);

	if (tkbi->cpufreq_req_h_freq)
		pm_qos_remove_request(tkbi->cpufreq_req_h_freq);

	if (tkbi->cpufreq_req_l_freq)
		pm_qos_remove_request(tkbi->cpufreq_req_l_freq);

	if (tkbi->thermal_req_h_freq)
		pm_qos_remove_request(tkbi->thermal_req_h_freq);

	if (tkbi->thermal_req_l_freq)
		pm_qos_remove_request(tkbi->thermal_req_l_freq);

	kfree(tkbi->cpufreq_req_l_freq);
	kfree(tkbi->cpufreq_req_h_freq);
	kfree(tkbi->thermal_req_l_freq);
	kfree(tkbi->thermal_req_h_freq);
	cancel_work_sync(&tkbi->cpufreq_work);
#ifdef CONFIG_SENSORS_CORETEMP_INTERRUPT
	cancel_work_sync(&tkbi->set_threshold);
#endif

	kfree(tkbi);

	misc_deregister(&tkb_misc);
}
module_exit(tkb_exit);

MODULE_LICENSE("GPL");
