/*
 *  Headset device detection driver.
 *
 * Copyright (C) 2011 ASUSTek Corporation.
 *
 * Authors:
 *  Jason Cheng <jason4_cheng@asus.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/irq.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/errno.h>
#include <linux/err.h>
#include <linux/switch.h>
#include <linux/slab.h>
#include <linux/wakelock.h>
#include <linux/gpio.h>
#include <asm/uaccess.h>
#include <asm/string.h>
#include <sound/soc.h>
#include <sound/jack.h>
#include "../codecs/rt5648.h"

MODULE_DESCRIPTION("Headset detection driver");
MODULE_LICENSE("GPL");


/*----------------------------------------------------------------------------
 ** FUNCTION DECLARATION
 **----------------------------------------------------------------------------*/
static int __init headset_init(void);
static void __exit headset_exit(void);
static int detection_work_fe375cg(void);
static int headset_get_jack_status(void);
static int headset_get_hook_status(void);
static int headset_disable_button_irq(void);
int hs_micbias_power(int on);
int hs_micbias_switch(int on);
int hs_get_status(void);
int hs_resume(void);
int hs_suspend(void);

/*----------------------------------------------------------------------------
 ** GLOBAL VARIABLES
 **----------------------------------------------------------------------------*/
enum {
	HS_JACK = 0,
};

enum {
	IRQ_HS_JACK = 0,
	IRQ_HS_HOOK,
	IRQ_HS_NONE,
};

struct snd_soc_jack_gpio fe375cg_hs_gpio[] = {
	[HS_JACK] = {
		.name			= "JACK_DET_AND_HOOK_DET",
		.gpio			= 14,
		.report			= SND_JACK_HEADSET |
					  SND_JACK_HEADPHONE |
					  SND_JACK_LINEOUT |
					  SND_JACK_BTN_0,
		.debounce_time		= 0,
		.jack_status_check	= detection_work_fe375cg,
		.irq_flags		= IRQF_TRIGGER_FALLING |
					    IRQF_TRIGGER_RISING,
	},
};
EXPORT_SYMBOL(fe375cg_hs_gpio);

static struct snd_soc_jack_gpio *jack_gpio;
static struct snd_soc_jack *hs_jack;

enum {
	OFF,
	ON,
};

struct headset_data {
	struct switch_dev sdev;
	struct wake_lock detection_lock;
	int status;
	bool enable_detect;
};
static struct headset_data hs_data;


extern struct snd_soc_codec *audio_codec;

int hs_get_status(void) {
	return hs_data.status;
}
EXPORT_SYMBOL(hs_get_status);

int hs_suspend(void)
{
	struct snd_soc_jack *jack = jack_gpio[HS_JACK].jack;
	hs_data.enable_detect = false;

	hs_jack = jack;
	hs_data.status = 0;
	headset_disable_button_irq();
	snd_soc_jack_free_gpios(jack, 1, &jack_gpio[HS_JACK]);

	return 0;
}
EXPORT_SYMBOL(hs_suspend);

int hs_resume(void)
{
	int ret = 0;

	ret = snd_soc_jack_add_gpios(hs_jack, 1, &jack_gpio[HS_JACK]);

	if (ret)
		pr_err("%s: add jack gpio failed\n", __func__);
	hs_data.enable_detect = true;

	return ret;

}
EXPORT_SYMBOL(hs_resume);

static int headset_get_jack_status(void)
{
	return rt5648_get_gpio_value(RT5648_GPIO_JD1_1);
}

static int headset_get_hook_status(void)
{
	return rt5648_get_gpio_value(RT5648_GPIO_6);
}

static int headset_enable_button_irq(void)
{
	int ret  = 0;

	ret = rt5648_enable_jd1_irq();
	ret = rt5648_set_jd1_irq_source(RT5648_GPIO_6);

	return ret;
}

static int headset_disable_button_irq(void)
{
	int ret = 0;

	ret = rt5648_disable_jd1_irq();
	ret = rt5648_set_jd1_irq_source(0);

	return ret;
}
static int headset_check_irq_source(void)
{
	int irq_source = IRQ_HS_JACK;

	if ((hs_data.status & SND_JACK_HEADSET) == SND_JACK_HEADSET) {
		if (rt5648_is_micbias_over_current()) {
			if (headset_get_jack_status())
				irq_source = IRQ_HS_JACK;
			else if (((hs_data.status & SND_JACK_BTN_0)
				    == SND_JACK_BTN_0) && (headset_get_hook_status() == 0))
				irq_source = IRQ_HS_HOOK;
			else if (headset_get_hook_status() == 1)
				irq_source = IRQ_HS_HOOK;
			else
				irq_source = IRQ_HS_NONE;
		} else {
			if (((hs_data.status & SND_JACK_BTN_0)
				    == SND_JACK_BTN_0) && (headset_get_hook_status() == 0))
				irq_source = IRQ_HS_HOOK;
			else if (headset_get_jack_status())
				irq_source = IRQ_HS_JACK;
			else
				irq_source = IRQ_HS_NONE;
		}
	}

	pr_debug("%s: irq_source = %d\n", __func__, irq_source);

	return irq_source;
}

static ssize_t headset_name_show(struct switch_dev *sdev, char *buf)
{
	const struct snd_soc_jack *jack = jack_gpio[HS_JACK].jack;
	if (jack) {
		printk("jack_type %d\n", jack->status);
		switch ((jack->status & jack_gpio[HS_JACK].report)) {
			case 0:
				return sprintf(buf, "NO DEVICE\n");
			case SND_JACK_HEADSET:
				return sprintf(buf, "HEADSET\n");
			case SND_JACK_HEADPHONE:
				return sprintf(buf, "HEADPHONE\n");
		}
	}
	return -EINVAL;
}

static ssize_t headset_state_show(struct switch_dev *sdev, char *buf)
{
	const struct snd_soc_jack *jack = jack_gpio[HS_JACK].jack;
	if (jack) {
		printk("jack_state %d\n", jack->status);
		switch ((jack->status & jack_gpio[HS_JACK].report)) {
			case 0:
				return sprintf(buf, "%d\n", 0);
			case SND_JACK_HEADSET:
				return sprintf(buf, "%d\n", 1);
			case SND_JACK_HEADPHONE:
				return sprintf(buf, "%d\n", 2);
		}
	}
	return -EINVAL;
}

int hs_micbias_switch(int status)
{
	int ret = -1;
	unsigned bits;
	struct snd_soc_codec *codec = audio_codec;

	if (codec == NULL)
		goto failed;

	pr_debug("HEADSET: Turn %s micbias power\n",
			(status == ON) ? "on" : "off");

	if (status == OFF &&
			codec->dapm.bias_level != SND_SOC_BIAS_OFF) {
		pr_debug("HEADSET: Codec turned it on, leave as was.");
		return 1;
	}

	bits = RT5648_PWR_BG | RT5648_PWR_VREF2 |
		RT5648_PWR_VREF1 | RT5648_PWR_MB;
	ret = snd_soc_update_bits(codec, RT5648_PWR_ANLG1, bits,
			status == ON ? bits : 0);
	if (ret < 0)
		goto failed;

	bits = RT5648_PWR_MB1;
	ret = snd_soc_update_bits(codec, RT5648_PWR_ANLG2, bits,
			status == ON ? bits : 0);
	if (ret < 0)
		goto failed;
	return ret;
failed:
	pr_err("HEADSET: Codec update failure!\n");
	return ret;
}
EXPORT_SYMBOL(hs_micbias_switch);


static int detection_work_fe375cg(void)
{
	int jk_status;
	int status = 0;
	int retries = 10;
	int irq_source;

	const int debouncing = 500;
	const int WAIT_CODEC_WAKE = 100;

	/*Avoid unnecessary headset detection after resuming codec*/
	if (!hs_data.enable_detect) {
		msleep(50);
		if(headset_get_jack_status()) {
			pr_info("%s: Avoid unnecessary detection\n", __func__);
			return status; /*Due to jack pin is high, return status as 0*/
		}
	}

	irq_source = headset_check_irq_source();

	status = hs_data.status;

	if (irq_source == IRQ_HS_NONE)
		return status;


	if (irq_source == IRQ_HS_HOOK) {
		if (headset_get_hook_status()) { // set btn bit.
			pr_info("HEADSET: Button Pressed\n");
			status |= SND_JACK_BTN_0;
		} else { // clear btn bit.
			pr_info("HEADSET: Button Releaseed\n");
			status &= ~SND_JACK_BTN_0;
		}
		hs_data.status = status;
		return status;
	}

	/* prevent system from falling asleep again as we sleep */
	wake_lock(&hs_data.detection_lock);
	while (audio_codec->suspended) {
		msleep(WAIT_CODEC_WAKE);
		if (--retries == 0) {
			pr_err("HEADSET: codec didn't waked in time\n");
			break;
		}
	}
	hs_micbias_switch(ON);

	/* Delay for hook status stable. */
	msleep(debouncing);
	jk_status = headset_get_jack_status();
	switch (jk_status) {
	case 0:
		if (headset_get_hook_status()) {
			pr_info("%s: headphone\n", __func__);
			status = SND_JACK_HEADPHONE;
		} else {
			pr_info("%s: headset\n", __func__);
			status = SND_JACK_HEADSET;
		}
		break;
	case 1:
		pr_info("%s: remove headset\n", __func__);
		status = 0;
		break;
	default:
		status = 0;
		break;
	}
	if ((status & SND_JACK_HEADSET) != SND_JACK_HEADSET) {
		hs_micbias_switch(OFF);
		headset_disable_button_irq();
	} else
		headset_enable_button_irq();

	hs_data.status = status;
	wake_unlock(&hs_data.detection_lock);

	return status;
}


/**********************************************************
 **  Function: Headset driver init function
 **  Parameter: none
 **  Return value: none
 **
 ************************************************************/
static int __init headset_init(void)
{
	int ret;
	printk(KERN_INFO "%s+ #####\n", __func__);

	wake_lock_init(&hs_data.detection_lock, WAKE_LOCK_SUSPEND, "hs_wakelock");

	hs_data.sdev.name = "h2w";
	hs_data.sdev.print_name = headset_name_show;
	hs_data.sdev.print_state = headset_state_show;
	hs_data.status = 0;
	hs_data.enable_detect = true;

	jack_gpio = fe375cg_hs_gpio;

	ret = switch_dev_register(&hs_data.sdev);
	if (ret < 0)
		pr_err("Warning: test node creation failed.");

	printk(KERN_INFO "%s- #####\n", __func__);
	return 0;
}

/**********************************************************
 **  Function: Headset driver exit function
 **  Parameter: none
 **  Return value: none
 **
 ************************************************************/
static void __exit headset_exit(void)
{
	switch_dev_unregister(&hs_data.sdev);
}

late_initcall(headset_init);
module_exit(headset_exit);
